<?php
/**
 * @file
 * Sajari Search API service implementation.
 */

class SearchApiSajariService extends SearchApiAbstractService {

  /**
   * An array of sajari engine connections by keyed by index name.
   *
   * @var \Sajari\Engine\EngineClient[]
   */
  protected $engines;

  /**
   * {@inheritdoc}
   */
  public function configurationForm(array $form, array &$form_state) {
    $form['sajari_company'] = [
      '#type' => 'textfield',
      '#title' => t('Sajari Company machine name'),
      '#description' => t('The company machine name is used for all requests'),
    ];
    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function indexItems(SearchApiIndex $index, array $items) {

    if (empty($items)) {
      return [];
    }

    $engine = $this->getEngine($index);
    $indexed_ids = [];
    foreach ($items as $id => $item) {
      $meta['item_id'] = $id;
      $doc_id = $index->datasource()->getSajariId($id, $index);
      $fulltext = '';

      foreach ($item as $key => $field) {
        // @TODO, don't special purpose the body field, maybe a rendered version
        // of the entity is better.
        if ($key === 'body:value') {
          $fulltext = $field['value'];
          continue;
        }
        $meta[$key] = $field['value'];
      }

      // If we don't have a Sajari Id then it must be a new item so try and add
      // the new item.
      if (empty($doc_id)) {
        try {
          $sajari_id = $engine->add([
            'q' => $fulltext,
            'meta' => $meta,
            // Don't flag as a duplicate if they have a different item_id.
            // @TODO, get this added to the Sajari API.
            // @see https://www.drupal.org/node/2502177
            'dupemetakeys' => 'item_id',
          ]);
          $indexed_ids[$sajari_id] = $id;
        }
        catch (Exception $e) {
          $this->logError(sprintf('Error adding item %s to Sajari', $id), $e->getMessage());
        }
      }
      else {
        try {
          $engine->replace(['id' => $doc_id, 'q' => $fulltext, 'meta' => $meta]);
          $indexed_ids[$doc_id] = $id;
        }
        catch (Exception $e) {
          $this->logError(sprintf('Error updating item %s to Sajari', $id), $e->getMessage());
        }
      }
    }

    // Return the ids of the items that we were able to index. These ids must
    // match up to the entity_ids in Drupal and they're only used for marking
    // the item as "unchanged" since the last update.
    return $indexed_ids;
  }

  /**
   * {@inheritdoc}
   */
  public function deleteItems($ids = 'all', SearchApiIndex $index = NULL) {

    // If the index is null then we clear every index for this server.
    if (is_null($index)) {
      foreach (search_api_index_load_multiple(FALSE, array('server' => $this->server->machine_name)) as $index) {
        $this->deleteItems('all', $index);
      }
    }

    $engine = $this->getEngine($index);
    $ids_to_delete = [];

    // The "all" key is a special case which deletes everything on the index.
    if ($ids === 'all') {
      try {

        // @TODO, this seems very wrong. Either, our datasource() is going to
        // need to provide a list of all id's from the index table or we need
        // to confirm with Sajari that this will always work?
        $response = $engine->search(['q' => '*', 'maxresults' => 100]);
        foreach ($response['results'] as $result) {
          $ids_to_delete[] = $result['docId'];
        }
      }
      catch (\Exception $e) {
        $this->logError('Error retrieving all results for "delete all" action.', $e->getMessage());
      }
    }
    // If we have an array of ids then map them to Sajari doc ids.
    elseif (is_array($ids)) {
      foreach ($ids as $id) {
        $sajari_id = $index->datasource()->getSajariId($id, $index);
        if (!empty($sajari_id)) {
          $ids_to_delete[] = $sajari_id;
        }
      }
    }

    // Handle the deleting of the ids.
    foreach ($ids_to_delete as $sajari_id) {
      try {
        if (!$engine->remove(['id' => $sajari_id])) {
          throw new SearchApiException(sprintf('Unable to delete %s', $sajari_id));
        }
      }
      catch (\Exception $e) {
        $this->logError(sprintf('Error deleting doc %s', $sajari_id), $e->getMessage());
        throw new SearchApiException(sprintf('Unable to delete %s', $sajari_id));
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public function search(SearchApiQueryInterface $query) {
    $engine = $this->getEngine($query->getIndex());

    // Set the max and offset to make the pagination work.
    $search_criteria = [
      'maxresults' => $query->getOption('limit'),
      'page' => $query->getOption('offset'),
    ];

    // Filter criteria filters on specific keys in the $meta data we indexed on
    // the document.
    $filter_string = $this->buildFilterString($query);
    if (!empty($filter_string)) {
      $search_criteria['filters'] = $filter_string;
    }

    // Sajari also supports a free text search upto 1000 words which we don't
    // validate here.
    $search_text = $this->buildFreeTextSearchString($query);
    if (!empty($search_text)) {
      $search_criteria['q'] = $search_text;
    }

    try {
      $response = $engine->search($search_criteria);
      return $this->parseSajariResponseIntoResults($response);
    }
    catch (Exception $e) {
      $this->logError('Error searching sajari.', $e->getMessage());
    }

    return array('result count' => 0);
  }

  /**
   * Translate the Search API filters into a Sajari query string.
   *
   * This method currently only supports basic = and != filters. See the docs
   * for more filters: https://www.sajari.com/api-documentation#search
   *
   * @param \SearchApiQueryInterface $query
   *   The Search API query object.
   *
   * @return string
   *   A valid Sajari search query string.
   */
  protected function buildFilterString(SearchApiQueryInterface $query) {

    $sajari_filter_string = '';
    foreach ($query->getFilter()->getFilters() as $filter) {
      foreach ($filter->getFilters() as $nested_filter) {
        list($key, $value, $operator) = $nested_filter;

        // Run the views operators through a map to Sajari operators.
        $operator = $this->getOperator($operator);
        $sajari_filter_string .= $operator . $key . ',' . $value . '|';
      }
    }

    // {operator}{key},{val}
    return rtrim($sajari_filter_string, '|');
  }

  /**
   * Gets the Sajari operator given the views operator.
   *
   * @param string $operator
   *   The Views operator.
   *
   * @return string
   *   The Sajari operated exception.
   *
   * @throws \InvalidArgumentException
   */
  protected function getOperator($operator) {
    // A map from views operators to sajari operators.
    $sajari_operators_map = array(
      '=' => '=',
      '!=' => '!=',
    );
    if (!isset($sajari_operators_map[$operator])) {
      throw new \InvalidArgumentException(sprintf('Unsupported views operator %s. Check the Sajari service integration.'));
    }

    return $sajari_operators_map[$operator];
  }

  /**
   * Builds a concatenated string from the free text search input in views.
   *
   * @param \SearchApiQueryInterface $query
   *   The Search API Query object.
   *
   * @return string
   *   A string to be used by Sajari for the free text search.
   */
  protected function buildFreeTextSearchString(SearchApiQueryInterface $query) {

    // Remove possible extra whitespace.
    $search_text = trim($query->getOriginalKeys());

    // If we have a valid search string use it, otherwise for now we return *.
    return empty($search_text) ? '*' : $search_text;
  }

  /**
   * Parse the Sajari response into a valid Search API results array.
   *
   * $result = [
   *   'results' => [
   *     $entity_id => ['docId' => '3-1', 'score' => 1, 'meta' => []],
   *   ],
   *   'result count' => 1,
   * ];
   *
   * @param array $response
   *   The Sajari response returned by their search endpoint.
   *
   * @return array
   *   A results array keyed by entity id.
   */
  protected function parseSajariResponseIntoResults(array $response) {
    $parsed_results = [];
    foreach ($response['results'] as $result) {
      // Key the results by the entity id.
      $item_id = $result['meta']['item_id'];
      $parsed_results['results'][$item_id] = $result;
    }

    return $parsed_results + ['result count' => $response['totalMatches']];
  }

  /**
   * Gets the Sajari engine.
   *
   * @param \SearchApiIndex $index
   *   The Search API Index.
   *
   * @return \Sajari\Engine\EngineClient
   *   The Sajari client.
   */
  protected function getEngine(SearchApiIndex $index) {
    $index_name = search_api_sajari_get_index_name($index->machine_name);
    if (!isset($this->engines[$index_name])) {
      $this->engines[$index_name] = search_api_sajari_get_instance($index_name);
    }
    return $this->engines[$index_name];
  }

  /**
   * Helper for logging errors.
   *
   * @param string $action_message
   *   The error message from the action we were performing.
   * @param string $error_message
   *   The error message from the exception.
   */
  protected function logError($action_message, $error_message) {
    watchdog('search_api_sajari', sprintf($action_message . ' Error message: %s', $error_message));
  }

}
